/* stdlib */
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

/* maxim */
#include "board.h"
#include "dma.h"
#include "mxc_delay.h"
#include "mxc_device.h"
#include "mxc_pins.h"
#include "nvic_table.h"
#include "spi.h"
#include "tmr.h"
#include "tmr_revb.h"
#include "uart.h"

/* project */
#include "MAX41470.h"

// #define LOG_ENABLE

#define APP_DATA_TMR MXC_TMR1
#define APP_DATA_TMR_IRQn TMR1_IRQn

uint32_t lastTimestamp = 0;

#define TICKS_50US 3000
#define TICKS_100US (2 * TICKS_50US)
#define TICKS_150US (3 * TICKS_50US)
#define TICKS_200US (4 * TICKS_50US)
#define TICKS_TOL 1000
#define TICKS_IN_TOL(ticks, expected, tol) (((ticks) > (expected) - (tol)) && ((ticks) < (expected) + (tol)))

typedef enum {
    State_Noise,
    State_Pulses,
    State_SyncWord,
    State_Data
} State;

int state = State_Noise;
int syncWordIndex = 0;
uint32_t syncWordDurations[] = {
    TICKS_50US,
    TICKS_100US,
    TICKS_50US,
    TICKS_100US,
    TICKS_100US,
    TICKS_200US,
    TICKS_50US,
    TICKS_50US,
    TICKS_100US,
    TICKS_50US,
    TICKS_50US,
    TICKS_50US,
    TICKS_50US,
    TICKS_150US,
    TICKS_150US,
    TICKS_50US,
    TICKS_200US,
};
int syncWordIndexCount = 17;
uint8_t bits[20 * 16 + 1];
uint8_t* bitsWr = bits;
uint8_t* bitsEnd = bits + sizeof(bits);
uint8_t dataVal = 0;

#ifdef LOG_ENABLE
char logBuffer[8000];
char* logWr = logBuffer;
char* logEnd = logBuffer + sizeof(logBuffer) - 1;

int logReq = 0;
int logReqDelayed = 0;
#endif

void PrintMessage() {
    uint8_t* bitsRd = bits + 1;

    for (size_t i = 0; i < 20; i++) {
        for (int j = 7; j >= 0; j--) {
            uint8_t b1 = *bitsRd++;
            uint8_t b2 = *bitsRd++;

            if (b1 == 0 && b2 == 1) {
                printf("0");
            } else if (b1 == 1 && b2 == 0) {
                printf("1");
            } else {
                printf("X");
            }
        }
    }
    printf("\r\n");

    /*
    printf("\nRAW: ");

for (uint8_t* br = bits; br < bitsWr; br++) {
    if (*br == 0) {
        printf("0");
    } else {
        printf("1");
    }
}
printf("\n");
    */
}

#ifdef LOG_ENABLE
void LogWrite(char ch) {
    if (logWr != logEnd) {
        *logWr++ = ch;
    }
}

void LogNum(int number) {
    char buff[10];

    snprintf(buff, sizeof(buff), "%d", number);

    char* buffRd = buff;
    while (*buffRd) {
        LogWrite(*buffRd++);
    }
}

void LogFlush() {
    logWr = logBuffer;
}

void LogPrint() {
    *logWr = '\0';
    printf("%s", logBuffer);
    logWr = logBuffer;
}
#endif

void Timer_IrqHandler(void) {
    // collect timestamp
    uint32_t durationSinceLast = MXC_TMR_GetCapture(APP_DATA_TMR);
    MXC_TMR1->cnt = 0;

    MXC_TMR_ClearFlags(APP_DATA_TMR);
    NVIC_ClearPendingIRQ(APP_DATA_TMR_IRQn);

#ifdef LOG_ENABLE
    if (state == State_Noise) {
        LogWrite('N');
    } else if (state == State_Pulses) {
        LogWrite('P');
    } else if (state == State_SyncWord) {
        LogWrite('S');
        LogNum(syncWordIndex);
    } else if (state == State_Data) {
        LogWrite('D');
        LogNum((int)(bitsWr - bits));
    }
#endif

    MXC_GPIO_OutSet(MXC_GPIO0, MXC_GPIO_PIN_31);
    if (state == State_Noise) {
#ifdef LOG_ENABLE
        if (!logReqDelayed) {
            LogFlush();
        }
#endif
        if (TICKS_IN_TOL(durationSinceLast, TICKS_50US, TICKS_TOL)) {
            state = State_Pulses;
        }
    } else if (state == State_Pulses) {
        if (TICKS_IN_TOL(durationSinceLast, TICKS_100US, TICKS_TOL)) {
            state = State_SyncWord;
            syncWordIndex = 0;
        } else if (TICKS_IN_TOL(durationSinceLast, TICKS_50US, TICKS_TOL)) {
            state = State_Pulses;
        } else {
            state = State_Noise;
        }
    } else if (state == State_SyncWord) {
        if (TICKS_IN_TOL(durationSinceLast, syncWordDurations[syncWordIndex], TICKS_TOL)) {
            syncWordIndex++;
            if (syncWordIndex == 17) {
                state = State_Data;
                bitsWr = bits;
                dataVal = 0;
            }
#ifdef LOG_ENABLE
            if (syncWordIndex == 5) {
                logReqDelayed = 1;
            }
#endif
        } else if (TICKS_IN_TOL(durationSinceLast, TICKS_50US, TICKS_TOL)) {
            state = State_Pulses;
        } else {
            state = State_Noise;
        }
    } else if (state == State_Data) {
        if (TICKS_IN_TOL(durationSinceLast, TICKS_50US, TICKS_TOL)) {
            *bitsWr++ = dataVal;
            if (bitsWr == bitsEnd) {
#ifdef LOG_ENABLE
                logReq = 1;
                LogWrite('X');
#endif
                PrintMessage();
                state = State_Noise;
            }
            dataVal = !dataVal;
        } else if (TICKS_IN_TOL(durationSinceLast, TICKS_100US, TICKS_TOL)) {
            *bitsWr++ = dataVal;
            if (bitsWr == bitsEnd) {
#ifdef LOG_ENABLE
                logReq = 1;
                LogWrite('Y');
#endif
                PrintMessage();
                state = State_Noise;
            } else {
                *bitsWr++ = dataVal;
                if (bitsWr == bitsEnd) {
#ifdef LOG_ENABLE
                    logReq = 1;
                    LogWrite('Z');
#endif
                    PrintMessage();
                    state = State_Noise;
                }
            }
            dataVal = !dataVal;
        } else {
            if (bitsWr + 1 == bitsEnd) {
                *bitsWr++ = dataVal;
#ifdef LOG_ENABLE
                logReq = 1;
                LogWrite('I');
#endif
                PrintMessage();
                state = State_Noise;
            } else {
#ifdef LOG_ENABLE
                logReq = 1;
                LogWrite('Q');
#endif
                state = State_Noise;
            }
        }
    }

#ifdef LOG_ENABLE
    if (state == State_Noise) {
        LogWrite('N');
    } else if (state == State_Pulses) {
        LogWrite('P');
    } else if (state == State_SyncWord) {
        LogWrite('S');
        LogNum(syncWordIndex);
    } else if (state == State_Data) {
        LogWrite('D');
        LogNum((int)(bitsWr - bits));
    }
    LogWrite('\r');
    LogWrite('\n');
#endif

    MXC_GPIO_OutClr(MXC_GPIO0, MXC_GPIO_PIN_31);
}

int main(void) {
    int iStatus;

    printf("\r\nInit begin\r\n");

    iStatus = MAX41470_Init();
    if (iStatus) {
        printf("MAX41470_Init failed with status code %d.", iStatus);
        while (1) {
        }
    }

    uint8_t regs14[] = {0x04};
    iStatus = MAX41470_WriteReg(0x14, regs14, sizeof(regs14));
    if (iStatus) {
        printf("MAX41470_WriteReg failed with status code %d when configuring registers 0x00 - 0x0E.", iStatus);
        while (1) {
        }
    }

    MXC_Delay(1000);

    //                  0     1     2     3     4     5     6     7     8     9     A     B     C     D     E
    //                                                                        FREQ  FREQ  FREQ  PREAM PREAM PREAM
    // recomemnd 868.3 {0x45, 0x26, 0x00, 0x1A, 0xB3, 0x0D, 0x0B, 0x1F, 0x00, 0x36, 0x3E, 0x66, 0x0F, 0x00, 0x00};
    // moje      869.5 {0x45, 0x26, 0x00, 0x1A, 0xB3, 0x0D, 0x0B, 0x1F, 0x00, 0x36, 0x51, 0x9A, 0x00, 0x00, 0x00};
    // moje      868.4:
    uint8_t regs00[] = {0x44, 0x3E, 0x00, 0x10, 0xB8, 0x10, 0x0B, 0x1F, 0x00, 0x36, 0x40, 0x00, 0x0F, 0x00, 0x00};
    iStatus = MAX41470_WriteReg(0x00, regs00, sizeof(regs00));
    if (iStatus) {
        printf("MAX41470_WriteReg failed with status code %d when configuring registers 0x00 - 0x0E.", iStatus);
        while (1) {
        }
    }

    uint8_t regs19[] = {0x93};
    iStatus = MAX41470_WriteReg(0x19, regs19, sizeof(regs19));
    if (iStatus) {
        printf("MAX41470_Init failed with status code %d when configuring register 0x19.", iStatus);
        while (1) {
        }
    }

    regs14[0] = 0x05;
    iStatus = MAX41470_WriteReg(0x14, regs14, sizeof(regs14));
    if (iStatus) {
        printf("MAX41470_WriteReg failed with status code %d when configuring registers 0x00 - 0x0E.", iStatus);
        while (1) {
        }
    }

    MXC_Delay(1000);

    mxc_gpio_cfg_t dbg;
    dbg.func = MXC_GPIO_FUNC_OUT;
    dbg.mask = MXC_GPIO_PIN_31;
    dbg.pad = MXC_GPIO_PAD_NONE;
    dbg.port = MXC_GPIO0;
    dbg.vssel = MXC_GPIO_VSSEL_VDDIOH;
    MXC_GPIO_Config(&dbg);

    mxc_tmr_cfg_t tmr;
    tmr.bitMode = TMR_BIT_MODE_32;
    tmr.clock = MXC_TMR_32M_CLK;
    tmr.cmp_cnt = 0xFFFFFFFF;
    tmr.mode = MXC_S_TMR_CTRL0_MODE_A_DUAL_EDGE;
    // tmr.mode = MXC_S_TMR_CTRL0_MODE_A_CAPTURE;
    tmr.pol = 0;
    tmr.pres = TMR_PRES_1;

    int tmrStatus = MXC_TMR_Init(APP_DATA_TMR, &tmr, 0);
    if (tmrStatus) __BKPT(0);

    MXC_GPIO_Config(&gpio_cfg_tmr1_MapB);
    MXC_TMR_RevB_ConfigGeneric((mxc_tmr_revb_regs_t*)APP_DATA_TMR, &tmr);

    MXC_GPIO_SetVSSEL(MAX41470_DATA_GPIO, MXC_GPIO_VSSEL_VDDIOH, MAX41470_DATA_PIN);
    MXC_GPIO_SetVSSEL(MXC_GPIO0, MXC_GPIO_VSSEL_VDDIOH, MXC_GPIO_PIN_20);

    MXC_NVIC_SetVector(APP_DATA_TMR_IRQn, Timer_IrqHandler);
    NVIC_SetPriority(APP_DATA_TMR_IRQn, 0);
    NVIC_ClearPendingIRQ(APP_DATA_TMR_IRQn);
    NVIC_EnableIRQ(APP_DATA_TMR_IRQn);
    MXC_TMR_EnableInt(APP_DATA_TMR);

    MXC_TMR_Start(APP_DATA_TMR);

    printf("\r\nInit ok\r\n");

    while (1) {
#ifdef LOG_ENABLE
        if (logReq) {
            logReq = 0;
            LogPrint();
        } else if (logReqDelayed) {
            for (int i = 0; i < 10000; i++) {
                if (logReq) {
                    logReqDelayed = 0;
                    logReq = 0;
                    LogPrint();
                }
                MXC_Delay(100);
            }
            if (logReqDelayed) {
                logReqDelayed = 0;
                logReq = 0;
                LogPrint();
            }
        }
#else
        MXC_Delay(100000);
#endif
    }
}
